from lmfit import Model
import numpy as np
from lmfit.model import ModelResult
import funcnodes as fn


@fn.NodeDecorator(
    "lmfit.fit",
    description="Fits a lmfit Model to data",
    outputs=[{"type": ModelResult, "name": "result"}],
)
def fit(
    x: np.ndarray, y: np.ndarray, model: Model, try_guess: bool = True
) -> ModelResult:
    x = np.asarray(x)
    y = np.asarray(y)
    params = model.make_params()
    if try_guess:
        try:
            params = model.guess(y, x=x)
        except Exception:
            pass

    fit_results = model.fit(y, params=params, x=x)

    return fit_results


@fn.NodeDecorator(
    "lmfit.fit_summary",
    description="Summarizes the fit results",
    outputs=[{"name": "summary"}],
)
def fit_summary(fit_results: ModelResult) -> dict:
    return fit_results.summary()


@fn.NodeDecorator(
    "lmfit.fit_report",
    description="Generates a report of the fit results",
    outputs=[{"name": "report"}],
)
def fit_report(fit_results: ModelResult) -> str:
    return fit_results.fit_report()


FIT_SHELF = fn.Shelf(
    nodes=[fit, fit_summary, fit_report],
    name="Fitting",
    description="Nodes for fitting lmfit models",
    subshelves=[],
)



from typing import Optional, Tuple
import funcnodes as fn
from lmfit.model import Model, Parameters
import numpy as np


def _update_param_entries(src: fn.NodeIO, result: str):
    node = src.node
    if not node:
        return
    model: Model = node.inputs["model"].value
    if not hasattr(model, "param_hints"):
        return

    if result not in model.param_hints:
        return
    v = model.param_hints[result]

    for k in ["value", "min", "max"]:
        if node.inputs[k].is_connected():
            continue
        if k in v:
            node.inputs[k].set_value(v[k], does_trigger=False)
        else:
            node.inputs[k].set_value(fn.NoValue, does_trigger=False)


@fn.NodeDecorator(
    "lmfit.update_model_param",
    description="Update a parameter of a lmfit Model",
    outputs=[{"type": Model, "name": "updated_model"}],
    default_io_options={
        "model": {
            "on": {
                "after_set_value": fn.decorator.update_other_io(
                    "param_name",
                    lambda x: list(x.param_names),
                )
            }
        },
        "param_name": {"on": {"after_set_value": _update_param_entries}},
    },
)
def update_model_param(
    model: Model,
    param_name: str,
    value: Optional[float] = None,
    min: Optional[float] = None,
    max: Optional[float] = None,
) -> Model:
    if param_name not in model.param_names:
        raise ValueError(f"Model has no parameter named {param_name}")

    updates = {}
    if value is not None:
        updates["value"] = value
    if min is not None:
        updates["min"] = min
    if max is not None:
        updates["max"] = max

    model.set_param_hint(param_name, **updates)

    return model


@fn.NodeDecorator(
    "lmfit.guess_model",
    description="Guess the parameters of a lmfit Model",
    outputs=[
        {"type": Model, "name": "updated_model"},
        {"type": Parameters, "name": "parameters"},
    ],
)
def guess_model(
    model: Model, y: np.ndarray, x: np.ndarray, update: bool = True
) -> Tuple[Model, Parameters]:
    guess_result = model.guess(y, x=x)

    if update:
        for param_name in guess_result:
            model.set_param_hint(param_name, value=guess_result[param_name].value)

    return model, guess_result


PARAMS_SHELF = fn.Shelf(
    nodes=[update_model_param, guess_model],
    name="Parameters",
    description="Nodes working with lmfit Model parameters",
    subshelves=[],
)



import funcnodes as fn
from typing import Tuple
from lmfit.model import Model

import numpy as np

try:
    import pandas as pd

    PANDAS_INSTALLED = True
except (ImportError, ModuleNotFoundError):
    PANDAS_INSTALLED = False


@fn.NodeDecorator(
    "lmfit.predict",
    description="Use a lmfit Model to fit data",
    outputs=[
        {"name": "prediction"},
        {"name": "components"},
    ],
)
def predict(
    x: np.ndarray,
    model: Model,
) -> Tuple[np.ndarray, np.ndarray]:
    results = [
        v for k, v in model.eval_components(x=x, params=model.make_params()).items()
    ]

    # merge results to an NxC array where N is the number of data points and C is the number of components#
    merged_results = np.vstack(results).T

    return merged_results.sum(axis=1), merged_results


if PANDAS_INSTALLED:

    @fn.NodeDecorator(
        "lmfit.predict_df",
        description="Use a lmfit Model to fit data and return a DataFrame",
        outputs=[
            {"name": "predictions"},
        ],
    )
    def predict_df(
        x: np.ndarray,
        model: Model,
    ) -> pd.DataFrame:
        data = {
            k: v
            for k, v in model.eval_components(x=x, params=model.make_params()).items()
        }
        if len(data) > 1:
            total = np.vstack([v for k, v in data.items()]).T.sum(axis=1)

            data = {
                model.name: total,
                **data,
            }

        return pd.DataFrame(data)


USEMODEL_SHELF = fn.Shelf(
    nodes=[
        predict,
    ]
    + ([predict_df] if PANDAS_INSTALLED else []),
    name="Use Model",
    description="Nodes for using lmfit models",
    subshelves=[],
)



import funcnodes as fn
from lmfit.model import ModelResult
import numpy as np
from typing import Optional

_nodes = []
try:
    import funcnodes_plotly as fn_plotly  # noqa: F401

    import plotly.graph_objects as go

    @fn.NodeDecorator(
        "lmfit.plot_results_plotly",
        description="Plots the results of a lmfit fit",
        outputs=[{"type": go.Figure, "name": "figure"}],
        default_render_options={
            "data": {
                "src": "figure",
            }
        },
    )
    def plot_results_plotly(
        fit_results: ModelResult,
        x: Optional[np.ndarray] = None,
        fig: Optional[go.Figure] = None,
    ) -> go.Figure:
        if x is None:
            x = fit_results.userkws["x"]

        comp_y = fit_results.eval_components(
            x=x,
        )

        if fig is None:
            fig = go.Figure()

        if fit_results.data is not None:
            fig.add_trace(
                go.Scatter(
                    x=fit_results.userkws["x"],
                    y=fit_results.data,
                    mode="markers" if len(fit_results.data) < 100 else "lines",
                    name="data",
                )
            )

        for name, data in comp_y.items():
            fig.add_trace(go.Scatter(x=x, y=data, mode="lines", name=name))

        fig.add_trace(go.Scatter(x=x, y=fit_results.best_fit, mode="lines", name="fit"))
        if fit_results.residual is not None:
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=-fit_results.residual,
                    mode="lines",
                    name="residual",
                )
            )

        return fig

    _nodes.append(plot_results_plotly)

except ImportError:
    pass


PLOT_SHELF = fn.Shelf(
    nodes=_nodes,
    name="Plotting",
    description="Nodes for plotting lmfit results",
    subshelves=[],
)



"""
here models are made
"""

import funcnodes as fn
from typing import Optional, Union, Literal, Tuple
from lmfit.model import CompositeModel, Model, ModelResult
from lmfit.models import (
    SplineModel,
    ExpressionModel,
    PolynomialModel,
    ThermalDistributionModel,
    StepModel,
    RectangleModel,
    ConstantModel,
    ComplexConstantModel,
    LinearModel,
    QuadraticModel,
    GaussianModel,
    LorentzianModel,
    SplitLorentzianModel,
    VoigtModel,
    PseudoVoigtModel,
    MoffatModel,
    Pearson4Model,
    Pearson7Model,
    StudentsTModel,
    BreitWignerModel,
    LognormalModel,
    DampedOscillatorModel,
    DampedHarmonicOscillatorModel,
    ExponentialGaussianModel,
    SkewedGaussianModel,
    SkewedVoigtModel,
    DoniachModel,
    PowerLawModel,
    ExponentialModel,
)
import numpy as np
from .utils import autoprefix, update_model_params, model_composit_train_create
from ._auto_model import (
    ConstantModel_node,
    ComplexConstantModel_node,
    LinearModel_node,
    QuadraticModel_node,
    GaussianModel_node,
    Gaussian2dModel_node,
    LorentzianModel_node,
    SplitLorentzianModel_node,
    VoigtModel_node,
    PseudoVoigtModel_node,
    MoffatModel_node,
    Pearson4Model_node,
    Pearson7Model_node,
    StudentsTModel_node,
    BreitWignerModel_node,
    LognormalModel_node,
    DampedOscillatorModel_node,
    DampedHarmonicOscillatorModel_node,
    ExponentialGaussianModel_node,
    SkewedGaussianModel_node,
    SkewedVoigtModel_node,
    DoniachModel_node,
    PowerLawModel_node,
    ExponentialModel_node,
)


# SplineModel,ExpressionModel,PolynomialModel,ThermalDistributionModel,StepModel,RectangleModel


@fn.NodeDecorator(
    "lmfit.SplineModel",
    description="Create or add a SplineModel model",
    outputs=[{"type": SplineModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "xknots_min": {"hidden": True},
        "xknots_max": {"hidden": True},
    },
)
def SplineModel_node(
    xknots: np.ndarray,
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    xknots_min: Optional[np.ndarray] = None,
    xknots_max: Optional[np.ndarray] = None,
) -> Model:
    xknots = np.asarray(xknots)
    xknots = xknots.flatten()
    if xknots_min is None:
        xknots_min = np.zeros_like(xknots) - np.inf
    if xknots_max is None:
        xknots_max = np.zeros_like(xknots) + np.inf

    if not xknots_min.shape == xknots.shape == xknots_max.shape:
        raise ValueError("xknots, xknots_min and xknots_max must have the same shape")

    prefix = autoprefix(SplineModel, composite)
    model = SplineModel(prefix=prefix, xknots=xknots)

    for i, xn in enumerate(xknots):
        name = f"{prefix}s{i}"
        model.set_param_hint(
            name,
            min=xknots_min[i],
            max=xknots_max[i],
        )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    "lmfit.ExpressionModel",
    description="Create or add a ExpressionModel model",
    outputs=[{"type": Model, "name": "model"}],
)
def ExpressionModel_node(
    expression: str,
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    independent_vars: Optional[Union[str, list[str]]] = None,
) -> Model:
    if independent_vars is not None:
        if isinstance(independent_vars, str):
            independent_vars = independent_vars.split(",")
        independent_vars = [s.strip() for s in independent_vars]
    # expresssion model has no prefix, the vars are as in the expression
    # prefix = autoprefix(ExpressionModel, composite)

    model = ExpressionModel(expression, independent_vars=independent_vars)

    # set all params to 1
    for param in model.param_names:
        model.set_param_hint(param, value=1)

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    "lmfit.PolynomialModel",
    description="Create or add a PolynomialModel model",
    outputs=[{"type": Model, "name": "model"}],
    default_io_options={
        "degree": {
            "value_options": {"max": 7, "min": 0},
        },
        "model": {"on": {"after_set_value": update_model_params()}},
        "min_c": {"hidden": True},
        "max_c": {"hidden": True},
    },
)
def PolynomialModel_node(
    degree: int,
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    min_c: float = -(10**8),
    max_c: float = 10**8,
) -> Model:
    prefix = autoprefix(PolynomialModel, composite)

    # clip degree to 7
    degree = int(min(7, max(0, degree)))
    model = PolynomialModel(degree, prefix=prefix)

    # default bounds is inf which does not work for polynomial models
    for i in range(degree + 1):
        model.set_param_hint(f"{prefix}c{i}", min=min_c, max=max_c)

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    "lmfit.ThermalDistributionModel",
    description="Create or add a ThermalDistributionModel model",
    outputs=[{"type": Model, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "kt_value": {"hidden": True},
        "kt_min": {"hidden": True},
        "kt_max": {"hidden": True},
        "kt_vary": {"hidden": True},
    },
)
def ThermalDistributionModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    form: Literal["bose", "maxwell", "fermi"] = "bose",
    amplitude_value: Optional[float] = 1,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: Optional[bool] = True,
    center_value: Optional[float] = 0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: Optional[bool] = True,
    kt_value: Optional[float] = 1,
    kt_min: Optional[float] = None,
    kt_max: Optional[float] = None,
    kt_vary: Optional[bool] = True,
) -> Model:
    prefix = autoprefix(ThermalDistributionModel, composite)
    model = ThermalDistributionModel(prefix=prefix, form=form)
    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )

    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )

    model.set_param_hint(
        "kt",
        value=kt_value,
        min=kt_min if kt_min is not None else -np.inf,
        max=kt_max if kt_max is not None else np.inf,
        vary=kt_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    "lmfit.StepModel",
    description="Create or add a StepModel model",
    outputs=[{"type": Model, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
    },
)
def StepModel_node(
    composite: Optional[CompositeModel] = None,
    y: Optional[np.ndarray] = None,
    x: Optional[np.ndarray] = None,
    form: Literal["linear", "atan", "erf", "logistic"] = "linear",
    amplitude_value: Optional[float] = 1,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: Optional[bool] = True,
    center_value: Optional[float] = 0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: Optional[bool] = True,
    sigma_value: Optional[float] = 1,
    sigma_min: Optional[float] = None,
    sigma_max: Optional[float] = None,
    sigma_vary: Optional[bool] = True,
) -> Model:
    prefix = autoprefix(StepModel, composite)
    model = StepModel(prefix=prefix, form=form)
    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )

    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )

    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    "lmfit.RectangleModel",
    description="Create or add a RectangleModel model",
    outputs=[{"type": Model, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center1_value": {"hidden": True},
        "center1_min": {"hidden": True},
        "center1_max": {"hidden": True},
        "center1_vary": {"hidden": True},
        "center2_value": {"hidden": True},
        "center2_min": {"hidden": True},
        "center2_max": {"hidden": True},
        "center2_vary": {"hidden": True},
        "sigma1_value": {"hidden": True},
        "sigma1_min": {"hidden": True},
        "sigma1_max": {"hidden": True},
        "sigma1_vary": {"hidden": True},
        "sigma2_value": {"hidden": True},
        "sigma2_min": {"hidden": True},
        "sigma2_max": {"hidden": True},
        "sigma2_vary": {"hidden": True},
    },
)
def RectangleModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    form: Literal["linear", "atan", "erf", "logistic"] = "linear",
    amplitude_value: Optional[float] = 1,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: Optional[bool] = True,
    center1_value: Optional[float] = 0,
    center1_min: Optional[float] = None,
    center1_max: Optional[float] = None,
    center1_vary: Optional[bool] = True,
    center2_value: Optional[float] = 1,
    center2_min: Optional[float] = None,
    center2_max: Optional[float] = None,
    center2_vary: Optional[bool] = True,
    sigma1_value: Optional[float] = 1,
    sigma1_min: Optional[float] = None,
    sigma1_max: Optional[float] = None,
    sigma1_vary: Optional[bool] = True,
    sigma2_value: Optional[float] = 1,
    sigma2_min: Optional[float] = None,
    sigma2_max: Optional[float] = None,
    sigma2_vary: Optional[bool] = True,
) -> Model:
    prefix = autoprefix(RectangleModel, composite)
    model = RectangleModel(prefix=prefix, form=form)
    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )

    model.set_param_hint(
        "center1",
        value=center1_value,
        min=center1_min if center1_min is not None else -np.inf,
        max=center1_max if center1_max is not None else np.inf,
        vary=center1_vary,
    )

    model.set_param_hint(
        "center2",
        value=center2_value,
        min=center2_min if center2_min is not None else -np.inf,
        max=center2_max if center2_max is not None else np.inf,
        vary=center2_vary,
    )

    model.set_param_hint(
        "sigma1",
        value=sigma1_value,
        min=sigma1_min if sigma1_min is not None else -np.inf,
        max=sigma1_max if sigma1_max is not None else np.inf,
        vary=sigma1_vary,
    )

    model.set_param_hint(
        "sigma2",
        value=sigma2_value,
        min=sigma2_min if sigma2_min is not None else -np.inf,
        max=sigma2_max if sigma2_max is not None else np.inf,
        vary=sigma2_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    "lmfit.model_merge",
    description="Merges two lmfit models with a defined operator",
    outputs=[{"name": "model"}],
)
def merge_models(
    a: Model, b: Model, operator: Literal["+", "-", "*", "/"] = "+"
) -> CompositeModel:
    if operator == "+":
        return a + b
    if operator == "-":
        return a - b
    if operator == "*":
        return a * b
    if operator == "/":
        return a / b
    raise ValueError(f"unknown operator: {operator}")


AUTOMODELS = [
    ConstantModel,
    ComplexConstantModel,
    LinearModel,
    QuadraticModel,
    GaussianModel,
    LorentzianModel,
    SplitLorentzianModel,
    VoigtModel,
    PseudoVoigtModel,
    MoffatModel,
    Pearson4Model,
    Pearson7Model,
    StudentsTModel,
    BreitWignerModel,
    LognormalModel,
    DampedOscillatorModel,
    DampedHarmonicOscillatorModel,
    ExponentialGaussianModel,
    SkewedGaussianModel,
    SkewedVoigtModel,
    DoniachModel,
    PowerLawModel,
    ExponentialModel,
]

AUTOMODELMAP = {model.__name__: model for model in AUTOMODELS}


@fn.NodeDecorator(
    "lmfit.auto_model",
    description="Automatically generate a model from the data",
    outputs=[{"type": Model, "name": "model"}, {"type": ModelResult, "name": "result"}],
    default_io_options={
        "r2_threshold": {
            "value": 0.95,
            "value_options": {"min": 0, "max": 1},
        },
    },
    seperate_thread=True,
)
def auto_model(
    x: np.ndarray, y: np.ndarray, r2_threshold: float = 0.95, iterations: int = 1
) -> Tuple[Model, ModelResult]:
    """
    Automatically generate a model from the data

    Parameters
    ----------
    x : np.ndarray
        x data
    y : np.ndarray
        y data
    r2_threshold : float, optional
        R^2 threshold for the model, by default 0.95
        Models will be generated until the R^2 is above this threshold or no more models can be generated.

    Returns
    -------
    Tuple[Model, ModelResult]
        The model and the fit result
    """

    r2_threshold = max(0, min(1, r2_threshold))

    composite_model = None
    composite_model_fit = None
    iterations = int(max(1, iterations))
    for i in range(iterations):
        best_model = None
        best_r2 = -np.inf

        if composite_model_fit is not None:
            y_res = y - composite_model_fit.eval(x=x)
        else:
            y_res = y

        for modelclass in AUTOMODELS:
            prefix = autoprefix(modelclass, composite_model)
            model: Model = modelclass(prefix=prefix)
            try:
                guess = model.guess(data=y_res, x=x)
            except Exception:
                guess = model.make_params()
            try:
                fit: ModelResult = model.fit(data=y_res, params=guess, x=x)
            except Exception:
                continue
            r2 = fit.rsquared
            if r2 > best_r2:
                # update model parameters to the best fit
                for param in fit.params:
                    model.set_param_hint(param, value=fit.params[param].value)

                best_r2 = r2
                best_model = model

            if best_r2 >= r2_threshold:
                break

        if composite_model is None:
            composite_model = best_model
        else:
            composite_model += best_model

        # final fit
        composite_model_fit = composite_model.fit(data=y, x=x)

        for param in composite_model_fit.params:
            composite_model.set_param_hint(
                param, value=composite_model_fit.params[param].value
            )

    return composite_model, composite_model_fit


@fn.NodeDecorator(
    "lmfit.quickmodel",
    description="Quick model creation",
    outputs=[{"type": Model, "name": "model"}],
    default_io_options={
        "modelname": {"value_options": {"options": list(AUTOMODELMAP.keys())}},
    },
)
def quickmodel(
    modelname: str,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    composite: Optional[CompositeModel] = None,
):
    """
    Quick model creation
    """
    modelclass = AUTOMODELMAP.get(modelname)
    if modelclass is None:
        raise ValueError(f"Unknown model: {modelname}")

    prefix = autoprefix(modelclass, composite)
    model = modelclass(prefix=prefix)

    return model_composit_train_create(model, composite, x, y)


MODEL_SHELF = fn.Shelf(
    name="Models",
    description="Lmfit models",
    nodes=[
        auto_model,
        quickmodel,
        ConstantModel_node,
        ComplexConstantModel_node,
        LinearModel_node,
        QuadraticModel_node,
        GaussianModel_node,
        Gaussian2dModel_node,
        LorentzianModel_node,
        SplitLorentzianModel_node,
        VoigtModel_node,
        PseudoVoigtModel_node,
        MoffatModel_node,
        Pearson4Model_node,
        Pearson7Model_node,
        StudentsTModel_node,
        BreitWignerModel_node,
        LognormalModel_node,
        DampedOscillatorModel_node,
        DampedHarmonicOscillatorModel_node,
        ExponentialGaussianModel_node,
        SkewedGaussianModel_node,
        SkewedVoigtModel_node,
        DoniachModel_node,
        PowerLawModel_node,
        ExponentialModel_node,
        SplineModel_node,
        ExpressionModel_node,
        PolynomialModel_node,
        ThermalDistributionModel_node,
        StepModel_node,
        RectangleModel_node,
    ],
    subshelves=[
        fn.Shelf(
            name="Model Operations",
            description="Operations with lmfit Models",
            nodes=[merge_models],
            subshelves=[],
        )
    ],
)



from typing import Type, Union, Optional
from lmfit.model import Model, CompositeModel
import funcnodes as fn
import numpy as np


def model_base_name(model: Union[Model, Type[Model]]) -> str:
    if isinstance(model, CompositeModel) or issubclass(model, CompositeModel):
        raise ValueError("CompositeModel is not supported")

    try:
        return model.func.__name__
    except Exception:
        pass

    if isinstance(model, Model):
        return model.__class__.__name__
    else:
        return model.__name__


def autoprefix(model: Type[Model], composite: Optional[CompositeModel] = None) -> str:
    basename = model_base_name(model)
    if composite is None:
        return f"{basename}_{1}_"

    prefix = 1
    prefixes = [p.prefix for p in composite.components]
    while f"{basename}_{prefix}_" in prefixes:
        prefix += 1
    return f"{basename}_{prefix}_"


def update_model_params():
    def func(src: fn.NodeIO, result: Model):
        node = src.node
        if not node:
            return
        if isinstance(result, CompositeModel):
            result = result.components[-1]
        params = result.make_params()
        for paramname, paramrootname in zip(
            result.param_names, result._param_root_names
        ):
            value = params[paramname].value
            try:
                node.inputs[paramrootname + "_value"].set_value(
                    value, does_trigger=False
                )
            except KeyError:
                pass

    return func


def model_composit_train_create(
    model: Model,
    composite: Optional[CompositeModel] = None,
    x: Optional[Union[list, np.ndarray]] = None,
    y: Optional[Union[list, np.ndarray]] = None,
) -> Model:
    if x is None and composite is not None:
        if hasattr(composite, "_current_x"):
            x = composite._current_x

    if y is None and composite is not None:
        if hasattr(composite, "_current_y"):
            y = composite._current_y

    if y is not None and x is not None:
        x = np.asarray(x)
        y = np.asarray(y)
        if x.shape[0] == y.shape[0]:
            if composite is not None:
                composite_model_fit = composite.fit(data=y, x=x)
                y_res = y - composite_model_fit.eval(x=x)
            else:
                y_res = y
            try:
                params = model.guess(y_res, x=x)
            except Exception:
                params = model.make_params()

            try:
                fit_res = model.fit(y_res, params=params, x=x)
                for param in fit_res.params:
                    model.set_param_hint(param, value=fit_res.params[param].value)
            except Exception:
                pass

    if composite is not None:
        model = composite + model

    model._current_x = x
    model._current_y = y
    return model



from lmfit.model import CompositeModel, Model
from .utils import autoprefix, update_model_params, model_composit_train_create
import numpy as np
import funcnodes as fn
from typing import Optional
from lmfit.models import (
    ConstantModel,
    ComplexConstantModel,
    LinearModel,
    QuadraticModel,
    GaussianModel,
    Gaussian2dModel,
    LorentzianModel,
    SplitLorentzianModel,
    VoigtModel,
    PseudoVoigtModel,
    MoffatModel,
    Pearson4Model,
    Pearson7Model,
    StudentsTModel,
    BreitWignerModel,
    LognormalModel,
    DampedOscillatorModel,
    DampedHarmonicOscillatorModel,
    ExponentialGaussianModel,
    SkewedGaussianModel,
    SkewedVoigtModel,
    DoniachModel,
    PowerLawModel,
    ExponentialModel,
)


@fn.NodeDecorator(
    id="lmfit.models.ConstantModel",
    description="Create or add a Constant model",
    outputs=[{"type": ConstantModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "c_value": {"hidden": True},
        "c_min": {"hidden": True},
        "c_max": {"hidden": True},
        "c_vary": {"hidden": True},
    },
)
def ConstantModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    c_value: float = 0.0,
    c_min: Optional[float] = None,
    c_max: Optional[float] = None,
    c_vary: bool = True,
) -> Model:
    prefix = autoprefix(ConstantModel, composite)
    model = ConstantModel(prefix=prefix)

    model.set_param_hint(
        "c",
        value=c_value,
        min=c_min if c_min is not None else -np.inf,
        max=c_max if c_max is not None else np.inf,
        vary=c_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.ComplexConstantModel",
    description="Create or add a Complex Constant model",
    outputs=[{"type": ComplexConstantModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "re_value": {"hidden": True},
        "re_min": {"hidden": True},
        "re_max": {"hidden": True},
        "re_vary": {"hidden": True},
        "im_value": {"hidden": True},
        "im_min": {"hidden": True},
        "im_max": {"hidden": True},
        "im_vary": {"hidden": True},
    },
)
def ComplexConstantModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    re_value: float = 0.0,
    re_min: Optional[float] = None,
    re_max: Optional[float] = None,
    re_vary: bool = True,
    im_value: float = 0.0,
    im_min: Optional[float] = None,
    im_max: Optional[float] = None,
    im_vary: bool = True,
) -> Model:
    prefix = autoprefix(ComplexConstantModel, composite)
    model = ComplexConstantModel(prefix=prefix)

    model.set_param_hint(
        "re",
        value=re_value,
        min=re_min if re_min is not None else -np.inf,
        max=re_max if re_max is not None else np.inf,
        vary=re_vary,
    )
    model.set_param_hint(
        "im",
        value=im_value,
        min=im_min if im_min is not None else -np.inf,
        max=im_max if im_max is not None else np.inf,
        vary=im_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.LinearModel",
    description="Create or add a Linear model",
    outputs=[{"type": LinearModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "slope_value": {"hidden": True},
        "slope_min": {"hidden": True},
        "slope_max": {"hidden": True},
        "slope_vary": {"hidden": True},
        "intercept_value": {"hidden": True},
        "intercept_min": {"hidden": True},
        "intercept_max": {"hidden": True},
        "intercept_vary": {"hidden": True},
    },
)
def LinearModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    slope_value: float = 1.0,
    slope_min: Optional[float] = None,
    slope_max: Optional[float] = None,
    slope_vary: bool = True,
    intercept_value: float = 0.0,
    intercept_min: Optional[float] = None,
    intercept_max: Optional[float] = None,
    intercept_vary: bool = True,
) -> Model:
    prefix = autoprefix(LinearModel, composite)
    model = LinearModel(prefix=prefix)

    model.set_param_hint(
        "slope",
        value=slope_value,
        min=slope_min if slope_min is not None else -np.inf,
        max=slope_max if slope_max is not None else np.inf,
        vary=slope_vary,
    )
    model.set_param_hint(
        "intercept",
        value=intercept_value,
        min=intercept_min if intercept_min is not None else -np.inf,
        max=intercept_max if intercept_max is not None else np.inf,
        vary=intercept_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.QuadraticModel",
    description="Create or add a Quadratic model",
    outputs=[{"type": QuadraticModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "a_value": {"hidden": True},
        "a_min": {"hidden": True},
        "a_max": {"hidden": True},
        "a_vary": {"hidden": True},
        "b_value": {"hidden": True},
        "b_min": {"hidden": True},
        "b_max": {"hidden": True},
        "b_vary": {"hidden": True},
        "c_value": {"hidden": True},
        "c_min": {"hidden": True},
        "c_max": {"hidden": True},
        "c_vary": {"hidden": True},
    },
)
def QuadraticModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    a_value: float = 0.0,
    a_min: Optional[float] = None,
    a_max: Optional[float] = None,
    a_vary: bool = True,
    b_value: float = 0.0,
    b_min: Optional[float] = None,
    b_max: Optional[float] = None,
    b_vary: bool = True,
    c_value: float = 0.0,
    c_min: Optional[float] = None,
    c_max: Optional[float] = None,
    c_vary: bool = True,
) -> Model:
    prefix = autoprefix(QuadraticModel, composite)
    model = QuadraticModel(prefix=prefix)

    model.set_param_hint(
        "a",
        value=a_value,
        min=a_min if a_min is not None else -np.inf,
        max=a_max if a_max is not None else np.inf,
        vary=a_vary,
    )
    model.set_param_hint(
        "b",
        value=b_value,
        min=b_min if b_min is not None else -np.inf,
        max=b_max if b_max is not None else np.inf,
        vary=b_vary,
    )
    model.set_param_hint(
        "c",
        value=c_value,
        min=c_min if c_min is not None else -np.inf,
        max=c_max if c_max is not None else np.inf,
        vary=c_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.GaussianModel",
    description="Create or add a Gaussian model",
    outputs=[{"type": GaussianModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
    },
)
def GaussianModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0.0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1.0,
    sigma_min: Optional[float] = 0,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
) -> Model:
    prefix = autoprefix(GaussianModel, composite)
    model = GaussianModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.Gaussian2dModel",
    description="Create or add a Gaussian-2D model",
    outputs=[{"type": Gaussian2dModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "centerx_value": {"hidden": True},
        "centerx_min": {"hidden": True},
        "centerx_max": {"hidden": True},
        "centerx_vary": {"hidden": True},
        "centery_value": {"hidden": True},
        "centery_min": {"hidden": True},
        "centery_max": {"hidden": True},
        "centery_vary": {"hidden": True},
        "sigmax_value": {"hidden": True},
        "sigmax_min": {"hidden": True},
        "sigmax_max": {"hidden": True},
        "sigmax_vary": {"hidden": True},
        "sigmay_value": {"hidden": True},
        "sigmay_min": {"hidden": True},
        "sigmay_max": {"hidden": True},
        "sigmay_vary": {"hidden": True},
    },
)
def Gaussian2dModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    centerx_value: float = 0.0,
    centerx_min: Optional[float] = None,
    centerx_max: Optional[float] = None,
    centerx_vary: bool = True,
    centery_value: float = 0.0,
    centery_min: Optional[float] = None,
    centery_max: Optional[float] = None,
    centery_vary: bool = True,
    sigmax_value: float = 1.0,
    sigmax_min: Optional[float] = 0,
    sigmax_max: Optional[float] = None,
    sigmax_vary: bool = True,
    sigmay_value: float = 1.0,
    sigmay_min: Optional[float] = 0,
    sigmay_max: Optional[float] = None,
    sigmay_vary: bool = True,
) -> Model:
    prefix = autoprefix(Gaussian2dModel, composite)
    model = Gaussian2dModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "centerx",
        value=centerx_value,
        min=centerx_min if centerx_min is not None else -np.inf,
        max=centerx_max if centerx_max is not None else np.inf,
        vary=centerx_vary,
    )
    model.set_param_hint(
        "centery",
        value=centery_value,
        min=centery_min if centery_min is not None else -np.inf,
        max=centery_max if centery_max is not None else np.inf,
        vary=centery_vary,
    )
    model.set_param_hint(
        "sigmax",
        value=sigmax_value,
        min=sigmax_min if sigmax_min is not None else -np.inf,
        max=sigmax_max if sigmax_max is not None else np.inf,
        vary=sigmax_vary,
    )
    model.set_param_hint(
        "sigmay",
        value=sigmay_value,
        min=sigmay_min if sigmay_min is not None else -np.inf,
        max=sigmay_max if sigmay_max is not None else np.inf,
        vary=sigmay_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.LorentzianModel",
    description="Create or add a Lorentzian model",
    outputs=[{"type": LorentzianModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
    },
)
def LorentzianModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0.0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1.0,
    sigma_min: Optional[float] = 0,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
) -> Model:
    prefix = autoprefix(LorentzianModel, composite)
    model = LorentzianModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.SplitLorentzianModel",
    description="Create or add a Split-Lorentzian model",
    outputs=[{"type": SplitLorentzianModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
        "sigma_r_value": {"hidden": True},
        "sigma_r_min": {"hidden": True},
        "sigma_r_max": {"hidden": True},
        "sigma_r_vary": {"hidden": True},
    },
)
def SplitLorentzianModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0.0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1.0,
    sigma_min: Optional[float] = 0,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
    sigma_r_value: float = 1.0,
    sigma_r_min: Optional[float] = 0,
    sigma_r_max: Optional[float] = None,
    sigma_r_vary: bool = True,
) -> Model:
    prefix = autoprefix(SplitLorentzianModel, composite)
    model = SplitLorentzianModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )
    model.set_param_hint(
        "sigma_r",
        value=sigma_r_value,
        min=sigma_r_min if sigma_r_min is not None else -np.inf,
        max=sigma_r_max if sigma_r_max is not None else np.inf,
        vary=sigma_r_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.VoigtModel",
    description="Create or add a Voigt model",
    outputs=[{"type": VoigtModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
    },
)
def VoigtModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0.0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1.0,
    sigma_min: Optional[float] = 0,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
) -> Model:
    prefix = autoprefix(VoigtModel, composite)
    model = VoigtModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.PseudoVoigtModel",
    description="Create or add a PseudoVoigt model",
    outputs=[{"type": PseudoVoigtModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
        "fraction_value": {"hidden": True},
        "fraction_min": {"hidden": True},
        "fraction_max": {"hidden": True},
        "fraction_vary": {"hidden": True},
    },
)
def PseudoVoigtModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0.0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1.0,
    sigma_min: Optional[float] = 0,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
    fraction_value: float = 0.5,
    fraction_min: Optional[float] = 0.0,
    fraction_max: Optional[float] = 1.0,
    fraction_vary: bool = True,
) -> Model:
    prefix = autoprefix(PseudoVoigtModel, composite)
    model = PseudoVoigtModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )
    model.set_param_hint(
        "fraction",
        value=fraction_value,
        min=fraction_min if fraction_min is not None else -np.inf,
        max=fraction_max if fraction_max is not None else np.inf,
        vary=fraction_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.MoffatModel",
    description="Create or add a Moffat model",
    outputs=[{"type": MoffatModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
        "beta_value": {"hidden": True},
        "beta_min": {"hidden": True},
        "beta_max": {"hidden": True},
        "beta_vary": {"hidden": True},
    },
)
def MoffatModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0.0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1,
    sigma_min: Optional[float] = 0,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
    beta_value: float = 1.0,
    beta_min: Optional[float] = None,
    beta_max: Optional[float] = None,
    beta_vary: bool = True,
) -> Model:
    prefix = autoprefix(MoffatModel, composite)
    model = MoffatModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )
    model.set_param_hint(
        "beta",
        value=beta_value,
        min=beta_min if beta_min is not None else -np.inf,
        max=beta_max if beta_max is not None else np.inf,
        vary=beta_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.Pearson4Model",
    description="Create or add a Pearson4 model",
    outputs=[{"type": Pearson4Model, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
        "expon_value": {"hidden": True},
        "expon_min": {"hidden": True},
        "expon_max": {"hidden": True},
        "expon_vary": {"hidden": True},
        "skew_value": {"hidden": True},
        "skew_min": {"hidden": True},
        "skew_max": {"hidden": True},
        "skew_vary": {"hidden": True},
    },
)
def Pearson4Model_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0.0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1.0,
    sigma_min: Optional[float] = None,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
    expon_value: float = 1.5,
    expon_min: Optional[float] = 0.500000000000001,
    expon_max: Optional[float] = 1000,
    expon_vary: bool = True,
    skew_value: float = 0.0,
    skew_min: Optional[float] = -1000,
    skew_max: Optional[float] = 1000,
    skew_vary: bool = True,
) -> Model:
    prefix = autoprefix(Pearson4Model, composite)
    model = Pearson4Model(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )
    model.set_param_hint(
        "expon",
        value=expon_value,
        min=expon_min if expon_min is not None else -np.inf,
        max=expon_max if expon_max is not None else np.inf,
        vary=expon_vary,
    )
    model.set_param_hint(
        "skew",
        value=skew_value,
        min=skew_min if skew_min is not None else -np.inf,
        max=skew_max if skew_max is not None else np.inf,
        vary=skew_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.Pearson7Model",
    description="Create or add a Pearson7 model",
    outputs=[{"type": Pearson7Model, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
        "expon_value": {"hidden": True},
        "expon_min": {"hidden": True},
        "expon_max": {"hidden": True},
        "expon_vary": {"hidden": True},
    },
)
def Pearson7Model_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0.0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1.0,
    sigma_min: Optional[float] = None,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
    expon_value: float = 1.5,
    expon_min: Optional[float] = None,
    expon_max: Optional[float] = 100,
    expon_vary: bool = True,
) -> Model:
    prefix = autoprefix(Pearson7Model, composite)
    model = Pearson7Model(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )
    model.set_param_hint(
        "expon",
        value=expon_value,
        min=expon_min if expon_min is not None else -np.inf,
        max=expon_max if expon_max is not None else np.inf,
        vary=expon_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.StudentsTModel",
    description="Create or add a StudentsT model",
    outputs=[{"type": StudentsTModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
    },
)
def StudentsTModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0.0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1.0,
    sigma_min: Optional[float] = 0.0,
    sigma_max: Optional[float] = 100,
    sigma_vary: bool = True,
) -> Model:
    prefix = autoprefix(StudentsTModel, composite)
    model = StudentsTModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.BreitWignerModel",
    description="Create or add a Breit-Wigner model",
    outputs=[{"type": BreitWignerModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
        "q_value": {"hidden": True},
        "q_min": {"hidden": True},
        "q_max": {"hidden": True},
        "q_vary": {"hidden": True},
    },
)
def BreitWignerModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0.0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1.0,
    sigma_min: Optional[float] = 0.0,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
    q_value: float = 1.0,
    q_min: Optional[float] = None,
    q_max: Optional[float] = None,
    q_vary: bool = True,
) -> Model:
    prefix = autoprefix(BreitWignerModel, composite)
    model = BreitWignerModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )
    model.set_param_hint(
        "q",
        value=q_value,
        min=q_min if q_min is not None else -np.inf,
        max=q_max if q_max is not None else np.inf,
        vary=q_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.LognormalModel",
    description="Create or add a Log-Normal model",
    outputs=[{"type": LognormalModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
    },
)
def LognormalModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0.0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1,
    sigma_min: Optional[float] = 0,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
) -> Model:
    prefix = autoprefix(LognormalModel, composite)
    model = LognormalModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.DampedOscillatorModel",
    description="Create or add a Damped Oscillator model",
    outputs=[{"type": DampedOscillatorModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
    },
)
def DampedOscillatorModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 1.0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 0.1,
    sigma_min: Optional[float] = 0,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
) -> Model:
    prefix = autoprefix(DampedOscillatorModel, composite)
    model = DampedOscillatorModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.DampedHarmonicOscillatorModel",
    description="Create or add a Damped Harmonic Oscillator model",
    outputs=[{"type": DampedHarmonicOscillatorModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
        "gamma_value": {"hidden": True},
        "gamma_min": {"hidden": True},
        "gamma_max": {"hidden": True},
        "gamma_vary": {"hidden": True},
    },
)
def DampedHarmonicOscillatorModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 1.0,
    center_min: Optional[float] = 0,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1.0,
    sigma_min: Optional[float] = 0,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
    gamma_value: float = 1.0,
    gamma_min: Optional[float] = 1e-19,
    gamma_max: Optional[float] = None,
    gamma_vary: bool = True,
) -> Model:
    prefix = autoprefix(DampedHarmonicOscillatorModel, composite)
    model = DampedHarmonicOscillatorModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )
    model.set_param_hint(
        "gamma",
        value=gamma_value,
        min=gamma_min if gamma_min is not None else -np.inf,
        max=gamma_max if gamma_max is not None else np.inf,
        vary=gamma_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.ExponentialGaussianModel",
    description="Create or add a Exponential Gaussian model",
    outputs=[{"type": ExponentialGaussianModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
        "gamma_value": {"hidden": True},
        "gamma_min": {"hidden": True},
        "gamma_max": {"hidden": True},
        "gamma_vary": {"hidden": True},
    },
)
def ExponentialGaussianModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1.0,
    sigma_min: Optional[float] = 0,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
    gamma_value: float = 1.0,
    gamma_min: Optional[float] = 0,
    gamma_max: Optional[float] = 20,
    gamma_vary: bool = True,
) -> Model:
    prefix = autoprefix(ExponentialGaussianModel, composite)
    model = ExponentialGaussianModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )
    model.set_param_hint(
        "gamma",
        value=gamma_value,
        min=gamma_min if gamma_min is not None else -np.inf,
        max=gamma_max if gamma_max is not None else np.inf,
        vary=gamma_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.SkewedGaussianModel",
    description="Create or add a Skewed Gaussian model",
    outputs=[{"type": SkewedGaussianModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
        "gamma_value": {"hidden": True},
        "gamma_min": {"hidden": True},
        "gamma_max": {"hidden": True},
        "gamma_vary": {"hidden": True},
    },
)
def SkewedGaussianModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0.0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1.0,
    sigma_min: Optional[float] = 0,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
    gamma_value: float = 0.0,
    gamma_min: Optional[float] = None,
    gamma_max: Optional[float] = None,
    gamma_vary: bool = True,
) -> Model:
    prefix = autoprefix(SkewedGaussianModel, composite)
    model = SkewedGaussianModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )
    model.set_param_hint(
        "gamma",
        value=gamma_value,
        min=gamma_min if gamma_min is not None else -np.inf,
        max=gamma_max if gamma_max is not None else np.inf,
        vary=gamma_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.SkewedVoigtModel",
    description="Create or add a Skewed Voigt model",
    outputs=[{"type": SkewedVoigtModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
        "skew_value": {"hidden": True},
        "skew_min": {"hidden": True},
        "skew_max": {"hidden": True},
        "skew_vary": {"hidden": True},
    },
)
def SkewedVoigtModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0.0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1.0,
    sigma_min: Optional[float] = 0,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
    skew_value: float = 0.0,
    skew_min: Optional[float] = None,
    skew_max: Optional[float] = None,
    skew_vary: bool = True,
) -> Model:
    prefix = autoprefix(SkewedVoigtModel, composite)
    model = SkewedVoigtModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )
    model.set_param_hint(
        "skew",
        value=skew_value,
        min=skew_min if skew_min is not None else -np.inf,
        max=skew_max if skew_max is not None else np.inf,
        vary=skew_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.DoniachModel",
    description="Create or add a Doniach model",
    outputs=[{"type": DoniachModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "center_value": {"hidden": True},
        "center_min": {"hidden": True},
        "center_max": {"hidden": True},
        "center_vary": {"hidden": True},
        "sigma_value": {"hidden": True},
        "sigma_min": {"hidden": True},
        "sigma_max": {"hidden": True},
        "sigma_vary": {"hidden": True},
        "gamma_value": {"hidden": True},
        "gamma_min": {"hidden": True},
        "gamma_max": {"hidden": True},
        "gamma_vary": {"hidden": True},
    },
)
def DoniachModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1.0,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    center_value: float = 0,
    center_min: Optional[float] = None,
    center_max: Optional[float] = None,
    center_vary: bool = True,
    sigma_value: float = 1.0,
    sigma_min: Optional[float] = None,
    sigma_max: Optional[float] = None,
    sigma_vary: bool = True,
    gamma_value: float = 0.0,
    gamma_min: Optional[float] = None,
    gamma_max: Optional[float] = None,
    gamma_vary: bool = True,
) -> Model:
    prefix = autoprefix(DoniachModel, composite)
    model = DoniachModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "center",
        value=center_value,
        min=center_min if center_min is not None else -np.inf,
        max=center_max if center_max is not None else np.inf,
        vary=center_vary,
    )
    model.set_param_hint(
        "sigma",
        value=sigma_value,
        min=sigma_min if sigma_min is not None else -np.inf,
        max=sigma_max if sigma_max is not None else np.inf,
        vary=sigma_vary,
    )
    model.set_param_hint(
        "gamma",
        value=gamma_value,
        min=gamma_min if gamma_min is not None else -np.inf,
        max=gamma_max if gamma_max is not None else np.inf,
        vary=gamma_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.PowerLawModel",
    description="Create or add a Power Law model",
    outputs=[{"type": PowerLawModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "exponent_value": {"hidden": True},
        "exponent_min": {"hidden": True},
        "exponent_max": {"hidden": True},
        "exponent_vary": {"hidden": True},
    },
)
def PowerLawModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    exponent_value: float = 1.0,
    exponent_min: Optional[float] = None,
    exponent_max: Optional[float] = None,
    exponent_vary: bool = True,
) -> Model:
    prefix = autoprefix(PowerLawModel, composite)
    model = PowerLawModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "exponent",
        value=exponent_value,
        min=exponent_min if exponent_min is not None else -np.inf,
        max=exponent_max if exponent_max is not None else np.inf,
        vary=exponent_vary,
    )

    return model_composit_train_create(model, composite, x, y)


@fn.NodeDecorator(
    id="lmfit.models.ExponentialModel",
    description="Create or add a Exponential model",
    outputs=[{"type": ExponentialModel, "name": "model"}],
    default_io_options={
        "model": {"on": {"after_set_value": update_model_params()}},
        "amplitude_value": {"hidden": True},
        "amplitude_min": {"hidden": True},
        "amplitude_max": {"hidden": True},
        "amplitude_vary": {"hidden": True},
        "decay_value": {"hidden": True},
        "decay_min": {"hidden": True},
        "decay_max": {"hidden": True},
        "decay_vary": {"hidden": True},
    },
)
def ExponentialModel_node(
    composite: Optional[CompositeModel] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    amplitude_value: float = 1,
    amplitude_min: Optional[float] = None,
    amplitude_max: Optional[float] = None,
    amplitude_vary: bool = True,
    decay_value: float = 1,
    decay_min: Optional[float] = None,
    decay_max: Optional[float] = None,
    decay_vary: bool = True,
) -> Model:
    prefix = autoprefix(ExponentialModel, composite)
    model = ExponentialModel(prefix=prefix)

    model.set_param_hint(
        "amplitude",
        value=amplitude_value,
        min=amplitude_min if amplitude_min is not None else -np.inf,
        max=amplitude_max if amplitude_max is not None else np.inf,
        vary=amplitude_vary,
    )
    model.set_param_hint(
        "decay",
        value=decay_value,
        min=decay_min if decay_min is not None else -np.inf,
        max=decay_max if decay_max is not None else np.inf,
        vary=decay_vary,
    )

    return model_composit_train_create(model, composite, x, y)


_AUTOMODELS = [
    ConstantModel_node,
    ComplexConstantModel_node,
    LinearModel_node,
    QuadraticModel_node,
    GaussianModel_node,
    Gaussian2dModel_node,
    LorentzianModel_node,
    SplitLorentzianModel_node,
    VoigtModel_node,
    PseudoVoigtModel_node,
    MoffatModel_node,
    Pearson4Model_node,
    Pearson7Model_node,
    StudentsTModel_node,
    BreitWignerModel_node,
    LognormalModel_node,
    DampedOscillatorModel_node,
    DampedHarmonicOscillatorModel_node,
    ExponentialGaussianModel_node,
    SkewedGaussianModel_node,
    SkewedVoigtModel_node,
    DoniachModel_node,
    PowerLawModel_node,
    ExponentialModel_node,
]



import funcnodes as fn
from .model import MODEL_SHELF
from .params import PARAMS_SHELF
from .plotting import PLOT_SHELF
from .fitting import FIT_SHELF
from .use_model import USEMODEL_SHELF

__version__ = "0.1.2"

NODE_SHELF = fn.Shelf(
    nodes=[],
    name="Funcnodes Lmfit",
    description="The nodes for the lmfit package",
    subshelves=[USEMODEL_SHELF, MODEL_SHELF, PARAMS_SHELF, PLOT_SHELF, FIT_SHELF],
)
